# -*- coding: utf-8 -*-
from django.conf import settings
from django.db.models.signals import class_prepared
import erppeek
import pprint
from django.db import models as djangomodels
from django.contrib.contenttypes.models import ContentType
from .fields import convert_field
from django.utils import translation

import sys
import inspect


def set_auth_cache():
    settings.CACHES = settings.CACHES or {}
    settings.CACHES["odoo_auth"] = {'BACKEND': 'django.core.cache.backends.memcached.MemcachedCache',
                                    'LOCATION': '127.0.0.1:18069'}


def set_odoo_client():
    config = getattr(settings, "ODOO_HOST", False)
    try:
        settings.odoo = erppeek.Client("%s:%d" % (config['HOST'], config['PORT']), db=config['DB'],
                                       user=config['USER'], password=config['PASSWORD'], verbose=False)
        settings.odoo.context = {"lang": settings.LANGUAGE_CODE}
        settings.odoo_models = {}
        settings.deferred_m2o = {}
        settings.deferred_o2m = {}
    except ConnectionRefusedError:
        print("Unable to connect to a running Odoo server.")
        raise


def add_extra_model_fields(sender, **kwargs):
    """Dynamically add the fields by reading the fields of the original ODOO model

        The fields are "translated" by using the definitions in fields
    """
    def add_field(django_model, field_details):
        odoo_field = convert_field(field_details)
        if odoo_field:
            field = odoo_field.to_django()
            field.contribute_to_class(django_model, field_details['name'])
#         else:
#             print field_details['name'], pprint.pformat(field_details)

    odoo = settings.odoo
    if getattr(sender, "_odoo_model", False):
        settings.odoo_models[sender._odoo_model] = sender
        _all_fields = odoo.model(sender._odoo_model).fields(sender._get_odoo_fields())
        for fname, fdetails in _all_fields.items():
            fdetails['name'] = fname
            fdetails['model'] = sender._odoo_model
            add_field(sender, fdetails)

        if sender._odoo_model in settings.deferred_m2o:
            for m2o_details in settings.deferred_m2o[sender._odoo_model]:
                origin = settings.odoo_models[m2o_details['model']]
                add_field(origin, m2o_details)
            settings.deferred_m2o[sender._odoo_model] = []

set_auth_cache()
set_odoo_client()
class_prepared.connect(add_extra_model_fields, dispatch_uid="FQFEQ#rfq3r")



# -*- coding: utf-8 -*-
from django.conf import settings
from django.db import models
from django.contrib.auth.models import User
from django.core.cache import caches
import erppeek

# TODO: traduction des DATA!!!
# TODO: lazy loading des objets many2one


class OdooModel(models.Model):

    """Model of a Odoo object copied in Django

        Attributes:
            _odoo_model: name of the Odoo model that will be copied in Django
            _odoo_fields: list of field names that will be copied from Odoo. If None, all field are copied.
            _odoo_ignore_fields: list of field names that will NOT be copied from Odoo
    """

    _odoo_model = None
    _odoo_fields = None
    _odoo_ignore_fields = None

    odoo_id = models.IntegerField(unique=True)

    def __init__(self, *args, **kwargs):
        self.translation_cache = {}
        return super(OdooModel, self).__init__(*args, **kwargs)

    class Meta:
        abstract = True

    @classmethod
    def _get_odoo_fields(cls):
        res = cls._odoo_fields or settings.odoo.model(cls._odoo_model).fields()
        return [f for f in res if not(f in (cls._odoo_ignore_fields or []))]

    @classmethod
    def odoo_load(cls, odoo_ids, client=None):
        """Loads records from Odoo

            Loads records from Odoo into Django instances given a list of Odoo identifiers *odoo_ids*.
            We read the data corresponding to the fields we need and convert it with respect
            to the type of field thanks to the methods defined in 'fields.py'. Each django field
            generated from a Odoo field contains a "odoo_field" attribute containing a "OdooField"
            instance.
        """
        def update_or_create(args):
            try:
                obj = cls.objects.get(odoo_id=args["odoo_id"])
                for (k, v) in args.items():
                    setattr(obj, k, v)
            except:
                obj = cls(**args)
                obj.save()
            return obj

        odoo_model = cls._odoo_model
        odoo_fields = cls._get_odoo_fields()
        client = client or settings.odoo
        records = client.model(odoo_model).read(odoo_ids, fields=odoo_fields, context=None)
        res = []
        for rec in records:
            args = {}
            args["odoo_id"] = rec["id"]
            del rec["id"]
            for field in cls._meta.fields:
                if hasattr(field, "odoo_field") and field.name in rec:
                    args[field.name] = field.odoo_field.convert_data(rec[field.name])

            res.append(update_or_create(args))
        return res

    @classmethod
    def odoo_search(cls, domain, offset=0, limit=None, order=None, context=None, client=None):
        """Search and load records from Odoo

            We load data from Odoo based on a domain filter
        """
        client = client or settings.odoo
        odoo_ids = client.search(cls._odoo_model, domain, offset=offset, limit=limit, order=order, context=context)
        return cls.odoo_load(odoo_ids, client=client) if odoo_ids else []

    @classmethod
    def odoo_write(cls, objs, args, client=None):
        """Writes in multiple records

            Writes the values provided in *args* into the Odoo records originating 
            the Django instances provided in *objs*
        """
        def convert(args):
            res = {}
            for field in cls._meta.fields:
                if field.name in args and hasattr(field, "odoo_field"):
                    res[field.name] = field.odoo_field.convert_back(args[field.name])
            return res

        client = client or settings.odoo
        odoo_model = cls._odoo_model
        odoo_ids = [o.odoo_id for o in objs if o.odoo_id]
        return client.model(odoo_model).write(odoo_ids, convert(args))

    @classmethod
    def cache_translation(cls, lang):
        """
            Récupère les traductions dans la langue `lang` des détails de tous les champs de l'objet
            Ces détails traduits sont stockés en cache dans chaque objet OdooField
        """
        def convert_lang(lang):
            print(lang)
            res = lang.replace("-", "_")
            if "_" in res:
                res = res[:3] + res[3:].upper()
            return res
        trans_fields = settings.odoo.execute(cls._odoo_model, 'fields_get', [], context={"lang": convert_lang(lang)})
        for field in cls._meta.fields:
            if hasattr(field, "odoo_field") and trans_fields.get(field.name):
                field.odoo_field.translation_cache[lang] = trans_fields[field.name]

    def _convert_to_push(self, fieldnames=None):
        res = {}
        fieldnames = fieldnames or type(self)._get_odoo_fields()
        for field in type(self)._meta.fields:
            if hasattr(field, "odoo_field") and field.name in fieldnames:
                res[field.name] = field.odoo_field.convert_back(getattr(self, field.name))
        return res

    def odoo_push(self, fieldnames=None, client=None):
        """Saves a Django instance into Odoo

            If the instance has an *odoo_id* then we call `write`, otherwise we call `create`; 
            we only save the values of the fields indicated in `fieldnames`, or all
            of them if it is None.

            :todo: deal with one2many and many2many fields?
        """
        odoo_model = type(self)._odoo_model
        client = client or settings.odoo
        args = self._convert_to_push(fieldnames)
        if self.odoo_id:
            client.model(odoo_model).write([self.odoo_id], args)
            return self.odoo_id
        else:
            return client.model(odoo_model).create(args)


class OdooUser(models.Model):
    user = models.OneToOneField(User, blank=False, related_name='odoo_user')

    def __init__(self, *args, **kwargs):
        config = getattr(settings, "ODOO_HOST", False)
        super(OdooUser, self).__init__(*args, **kwargs)
        passwd = kwargs.get('password') or caches["odoo_auth"].get('%s_credentials' % self.user.username)
        self.odoo_client = erppeek.Client("%s:%d" % (config['HOST'], config['PORT']), db=config['DB'],
                                          user=self.user.username, password=passwd, verbose=False)



# -*- coding: utf-8 -*-
from django.conf import settings
from django.db import models as djangomodels
from django.utils import translation
from django.utils.functional import lazy
from django.utils import six
from .models import OdooModel
import base64

"""
    Translation methods
"""


def _get_details_in_lang(field, lang):
    if field.translation_cache.get(lang):
        return field.translation_cache[lang]
    else:
        settings.odoo_models[field.details['model']].cache_translation(lang)
        try:
            return field.translation_cache[lang]
        except:
            return field.details


def field_translate(field, key):
    lang = translation.get_language()
    details = _get_details_in_lang(field, lang)
    res = details.get(key, "")
    if isinstance(res, six.binary_type):
        res = six.text_type(res)
    return res

_ = lazy(field_translate, six.text_type)


def selection_translate(field):
    def trans(val):
        lang = translation.get_language()
        details = _get_details_in_lang(field, lang)
        return dict(details['selection'])[val]

    trans_lazy = lazy(trans, six.text_type)

    res = []
    for val, _label in field.details.get('selection'):
        res.append((val, trans_lazy(val)))
    return tuple(res)


# TODO: default values
# TODO: domains

FIELDS_CONV = {
    "char": "CharField",
    "boolean": "BooleanField",
    "integer": "IntegerField",
    "text": "TextField",
    "float": "DecimalField",
    "date": "DateField",
    "datetime": "DateTimeField",
    "time": "TimeField",
    "binary": "BinaryField",
    "selection": "CharField",
    "many2one": "ForeignKey",
    "one2many": None,
    #     "many2many": "",
    #     "function": "",
    #     "related": "",
}


class OdooField(object):

    def __init__(self, details):
        self.details = details
        self.django_field = False
        self.translation_cache = {}  # translations cache
        return super(OdooField, self).__init__()

    def to_django(self, **kwargs):
        kwargs.update({
            "verbose_name": _(self, 'string'),
            "help_text": _(self, 'help'),
            "blank": not(self.details.get("required")),
            "editable": not(self.details.get("readonly")),
        })
        django_field = getattr(djangomodels, FIELDS_CONV[self.details["type"]])(**kwargs)
        django_field.odoo_field = self
        self.django_field = django_field
        return django_field

    def convert_data(self, data):
        return data or None

    def convert_back(self, data):
        return data or False


class TextField(OdooField):

    def to_django(self, **kwargs):
        if self.details.get("required"):
            kwargs["default"] = ""
        kwargs["null"] = not(self.details.get("required"))
        return super(TextField, self).to_django(**kwargs)


class CharField(TextField):

    def to_django(self, **kwargs):
        kwargs['max_length'] = self.details.get('size') or 512
        return super(CharField, self).to_django(**kwargs)


class BooleanField(OdooField):

    def to_django(self, **kwargs):
        kwargs["default"] = False
        return super(BooleanField, self).to_django(**kwargs)

    def convert_data(self, data):
        return data or False


class IntegerField(OdooField):

    def to_django(self, **kwargs):
        if self.details.get("required"):
            kwargs["default"] = 0
        kwargs["null"] = not(self.details.get("required"))
        return super(IntegerField, self).to_django(**kwargs)


class FloatField(IntegerField):

    def to_django(self, **kwargs):
        if self.details.get("digits"):
            kwargs["max_digits"] = self.details["digits"][0]
            kwargs["decimal_places"] = self.details["digits"][1]
        kwargs["null"] = not(self.details.get("required"))
        return super(FloatField, self).to_django(**kwargs)


class DateField(OdooField):

    def to_django(self, **kwargs):
        kwargs["null"] = not(self.details.get("required"))
        if self.details.get("required"):
            kwargs["auto_now_add"] = True
        return super(DateField, self).to_django(**kwargs)


class DateTimeField(DateField):
    pass


class TimeField(DateField):
    pass


class BinaryField(OdooField):

    def to_django(self, **kwargs):
        kwargs["null"] = not(self.details.get("required"))
        return super(BinaryField, self).to_django(**kwargs)

    def convert_data(self, data):
        """Odoo data is a b64-encoded string"""
        return base64.b64decode(data) if data else None

    def convert_back(self, data):
        return base64.b64encode(data).decode("utf-8") if data else False


class SelectionField(CharField):

    def to_django(self, **kwargs):
        kwargs["choices"] = selection_translate(self)
        return super(SelectionField, self).to_django(**kwargs)


class Many2OneField(OdooField):

    """
        If the model identified by details['relation'] exists in django, then we can create the field directly.
        Otherwise, we delay the field creation until the possible creation of this model.
    """

    def __new__(cls, details):
        if details['relation'] in settings.odoo_models:
            return OdooField.__new__(cls)
        else:
            settings.deferred_m2o[details['relation']] = settings.deferred_m2o.get(details['relation'], [])
            settings.deferred_m2o[details['relation']].append(details)
            return None

    def to_django(self, **kwargs):
        kwargs["null"] = not(self.details.get("required"))
        if self.details['relation'] == self.details['model']:
            kwargs["to"] = "self"
        else:
            to_model = settings.odoo_models[self.details['relation']]
            kwargs["to"] = to_model
        kwargs["related_name"] = ("%(model)s_%(name)s" % (self.details)).replace('.', '_')
        return super(Many2OneField, self).to_django(**kwargs)

    def convert_data(self, data):
        """
            Odoo data is a pair (id, label)
            We look for objects in the target model an instance having a odoo_id equal to the first
            element of the pair ; if not found, we load it from Odoo

            :param (tuple or False) data: the value to convert
            :return (OdooModel or False): the object instance linked to this m2o field
        """
        if data and isinstance(data, (list, tuple)) and len(data) == 2:
            to_model = settings.odoo_models[self.details['relation']]
            targets = to_model.objects.filter(odoo_id=data[0])
            if targets:
                return targets[0]
            else:
                return to_model.odoo_load([data[0]])[0]
        return data or None

    def convert_back(self, data):
        """
            Django data is either None or a Django instance
            We tranform it into False or an integer by getting the odoo_id on the instance.

            :todo: if the target objet has no odoo_id, first push it to odoo
            :param (OdooModel or False) data: the value to convert
            :return (integer or False): the idi of the object in odoo
        """
        if data and isinstance(data, OdooModel) and hasattr(data, 'odoo_id'):
            return data.odoo_id
#         elif isinstance(data, (int, long)):
#             return data
        else:
            return False


class One2ManyField(OdooField):

    """
        There is no one2many field in Django, so we simply set the "relation_field"
        attribute of the foreignKey field encoding the opposite relationship so it bares
        the name of this one2many field
    """
    def __new__(cls, details):
        if details['relation'] in settings.odoo_models:
            relation = settings.odoo_models[details['relation']]
            for field in relation._meta.Fields:
                if field.name == details['relation_field']:
                    field.related_name = details['name']
        else:
            settings.deferred_o2m[details['relation']] = settings.deferred_o2m.get(details['relation'], [])
            settings.deferred_o2m[details['relation']].append(details)


def convert_field(details):
    if not(details['type'] in FIELDS_CONV):
        return None
    return eval(details["type"].title() + "Field")(details)



from django.conf import settings
from django.contrib.auth.models import User
import erppeek
from .models import OdooUser
from django.core.cache import caches
from django.db import transaction


class OdooAuthBackend(object):

    """
    Authenticate against the user in Odoo
    """
    @transaction.atomic
    def authenticate(self, username=None, password=None):
        config = getattr(settings, "ODOO_HOST", False)
        try:
            odoo_client = erppeek.Client("%s:%d" % (config['HOST'], config['PORT']), db=config['DB'],
                                         user=username, password=password, verbose=False)
        except:
            return None

        caches["odoo_auth"].set('%s_credentials' % username, password, None)

        try:
            user = User.objects.get(username=username)
            odoo_user = user.odoo_user
        except User.DoesNotExist:
            # Create a new user. Note that we can set password
            # to anything, because it won't be checked; the password
            # from Odoo will.
            user = User(username=username, password='get from Odoo')
            user.is_staff = False
            user.is_superuser = False
            user.save()
            odoo_user = OdooUser(user=user)
            odoo_user.save()
        odoo_user.odoo_client = odoo_client
        return user

    def get_user(self, user_id):
        try:
            return User.objects.get(pk=user_id)
        except User.DoesNotExist:
            return None
